#include <View.h>
#include <Bitmap.h>
#include <Alert.h>
#include <TranslationUtils.h>
#include <RadioButton.h>
#include <Box.h>
#include <Message.h>
#include <StringView.h>
#include <Button.h>
#include <stdio.h>
#include <stdlib.h>
#include "consts.h"
#include "view.h"

view::view(BRect rect)
	: BView(rect, "", B_FOLLOW_ALL, B_WILL_DRAW | B_PULSE_NEEDED),
	minmove(MINMOVE_1),
	maxmove(MAXMOVE_1)
{
	SetFontSize(20);
	SetViewColor(233, 233, 233);
	image = BTranslationUtils::GetBitmap("bitmap2.bmp");
	
	if(image) // uhh
	if(((image->Bounds().IntegerWidth() + 1) < 192) || ((image->Bounds().IntegerHeight() + 1) < 192)) {
		delete image;
		image = 0;
	}
	if(!image) {
		image = new BBitmap(BRect(0, 0, 191, 191), B_RGB32);
		create_image(image);
	}
	AddChild(new BStringView(BRect(40, 10, 100, 30), "puz", "Puzzle:"));
	AddChild(new BStringView(BRect(250, 10, 300, 30), "tar", "Target:"));
	BBox *box = new BBox(BRect(35, 240, 220, 295));
	box->SetLabel("Difficulty:");
	box->SetFont(be_plain_font);
	AddChild(box);
	
	easy = new BRadioButton(BRect(5, 10, 80, 25), "easy", "Easy", new BMessage(MSGMOVE1));
	normal = new BRadioButton(BRect(5, 30, 80, 45), "norm", "Normal", new BMessage(MSGMOVE2));
	hard = new BRadioButton(BRect(100, 20, 175, 35), "hard", "Hard", new BMessage(MSGMOVE3));
	hard->SetValue(B_CONTROL_ON);
	box->AddChild(easy);
	box->AddChild(normal);
	box->AddChild(hard);
	
	newgame = new BButton(BRect(260, 240, 430, 295), "button", "New Game", new BMessage(NEWGAME));
	AddChild(newgame);
	for(int32 i = 0; i < 16; i++)
	{
		bmp[i] = new BBitmap(BRect(0, 0, 47, 47), B_RGB32);
		unpack(i);
	}
	reset();
}

void view::AttachedToWindow()
{
	easy->SetTarget(this);
	normal->SetTarget(this);
	hard->SetTarget(this);
	newgame->SetTarget(this);
	BView::AttachedToWindow();
}

void view::reset()
{
	srand(time(0));
	char ch = 0;
	for(int32 j = 0; j < 4; j++)
	for(int32 i = 0; i < 4; i++) {
		rects[j][i] = MRect(i * 48, j * 48, ((i * 48) + 48)-1, ((j * 48) + 48)-1);
		rects[j][i].OffsetBy(40, 40);
		rects[j][i].c = ch;
		rects[j][i].num = numbers[ch];
		matrix[j][i] = 1;
		ch++;
	}
	
	pos.j = 3;
	pos.i = 3;
	matrix[3][3] = 0;
	for(int32 i = 1; i < 16; i++) {
		int32 j = rand() % 16;
		shuffle((j >> 2) & 3, j & 3);
	}
}

void view::unpack(int32 i)
{
	int xp = 48 * (i % 4);
	int yp = 48 * (i / 4);

	int32 srow = image->BytesPerRow() >> 2;
	int32 drow = bmp[i]->BytesPerRow() >> 2;
	uint32 *src = (uint32 *)image->Bits() + yp * srow + xp;
	uint32 *dst = (uint32 *)bmp[i]->Bits();
	
	for(int32 y = 0; y < 48; y++)
	{
		for(int32 x = 0; x < 48; x++)
		{
			dst[x] = src[x];
		}
		src += srow;
		dst += drow;
	}
}

void view::shuffle(int32 i, int32 j)
{
	if(matrix[j][i] != 0) {
		int y = rand() & i;
		int x = rand() & j;
		char tempc = rects[y][x].c;
		int32 tempn = rects[y][x].num;
		if(tempc != rects[j][i].c && tempn != rects[j][i].num) {
			rects[y][x].c = rects[j][i].c;
			rects[y][x].num = rects[j][i].num;
			rects[j][i].c = tempc;
			rects[j][i].num = tempn;
		}
	}
}

bool view::winner()
{
	static char temp[64];
	int32 k = 0;
	for(int32 j = 0; j < 4; j++)
	for(int32 i = 0; i < 4; i++) {
		temp[k] = rects[j][i].c;
		k++;
	}
	if(memcmp(temp, result, 16) == 0) return true;
	return false;
}

void view::create_image(BBitmap *bitmap)
{
	uint32 *bits = (uint32 *)bitmap->Bits();
	int32 srow = bitmap->BytesPerRow() >> 2;
	
	for(int32 i = 0; i < 16; i++) 
	{
		int32 xp = 48 * (i % 4);
		int32 yp = 48 * (i / 4);
		uint32 *src = (uint32 *)bits + yp * srow + xp;
		for(int32 y = 0; y < 48; y++)
		{
			for(int32 x = 0; x < 48; x++)
			{
				src[x] = colours[i];
			}
			src += srow;
		}
	}
}

void view::MessageReceived(BMessage *m)
{
	switch(m->what)
	{
		case MSGMOVE1:
			minmove = MINMOVE_3;
			maxmove = MAXMOVE_3;
			break;
		case MSGMOVE2:
			minmove = MINMOVE_2;
			maxmove = MAXMOVE_2;
			break;
		case MSGMOVE3:
			minmove = MINMOVE_1;
			maxmove = MAXMOVE_1;
			break;
		case NEWGAME:
			reset();
			break;
		default:
			BView::MessageReceived(m);
			break;
	}
}

void view::Pulse()
{
	if(LockLooper()) {
		Draw(Bounds());
		UnlockLooper();
	}
}

void view::Draw(BRect rect)
{
	SetHighColor(222, 222, 222);
	FillRect(rects[pos.j][pos.i]);
	SetHighColor(0, 0, 0);
	
	for(int32 j = 0; j < 4; j++)
	for(int32 i = 0; i < 4; i++) {
		if(matrix[j][i]) {
			DrawBitmapAsync(bmp[rects[j][i].c], rects[j][i]);
			StrokeRect(rects[j][i]);
			DrawString(reinterpret_cast<const char *>(&rects[j][i].num),
				BPoint(rects[j][i].left + 20, rects[j][i].top + 30));
		}
	}
	
	DrawBitmap(image, BRect(0, 0, 191, 191), BRect(250, 40, 441, 231));
	BRect r(35, 35, 236, 236);
	StrokeRect(r);
	
	SetHighColor(255, 255, 255);
	r.InsetBy(1, 1);
	StrokeRect(r);
	
	SetHighColor(0, 0, 0);
	r.InsetBy(-1, -1);
	r.OffsetBy(210, 0);
	StrokeRect(r);
	
	SetHighColor(255, 255, 255);
	r.InsetBy(1, 1);
	StrokeRect(r);
	
	Sync();
	
	if(winner()) {
		(new BAlert("shuffle", "Puzzle complete", "Ok"))->Go();
		reset();
	}
}

void view::solve(int32 i, int32 j)
{
	int32 jj = j - pos.j;
	int32 ii = i - pos.i;
	
	if(pos.j != j && pos.i != i) return;
	if(jj >= minmove && jj <= maxmove && ii >= minmove && ii <= maxmove) {
		char tempc = rects[pos.j][pos.i].c;
		int32 tempn = rects[pos.j][pos.i].num;
		rects[pos.j][pos.i].c = rects[j][i].c;
		rects[pos.j][pos.i].num = rects[j][i].num;
		rects[j][i].c = tempc;
		rects[j][i].num = tempn;
		matrix[pos.j][pos.i] = 1;
		matrix[j][i] = 0;
		pos.j = j; pos.i = i;
	}
}

void view::MouseDown(BPoint where)
{
	for(int32 j = 0; j < 4; j++)
	for(int32 i = 0; i < 4; i++)
	{
		if(rects[j][i].Contains(where))
		{
			if(matrix[j][i] != 0)
			{
				solve(i, j);
				break;
			}
		}
	}
	BView::MouseDown(where);	
}
